@@ -15,7 +15,7 @@
 #  specific language governing permissions and limitations
 #  under the License.
 
-from .utils import NamespacedClient, query_params, _make_path, SKIP_IN_PATH
+from .utils import SKIP_IN_PATH, NamespacedClient, _make_path, query_params
 
 
 class EnrichClient(NamespacedClient):
@@ -24,7 +24,7 @@ class EnrichClient(NamespacedClient):
         """
         Deletes an existing enrich policy and its enrich index.
 
-        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.11/delete-enrich-policy-api.html>`_
+        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.12/delete-enrich-policy-api.html>`_
 
         :arg name: The name of the enrich policy
         """
@@ -43,7 +43,7 @@ class EnrichClient(NamespacedClient):
         """
         Creates the enrich index for an existing enrich policy.
 
-        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.11/execute-enrich-policy-api.html>`_
+        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.12/execute-enrich-policy-api.html>`_
 
         :arg name: The name of the enrich policy
         :arg wait_for_completion: Should the request should block until
@@ -64,7 +64,7 @@ class EnrichClient(NamespacedClient):
         """
         Gets information about an enrich policy.
 
-        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.11/get-enrich-policy-api.html>`_
+        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.12/get-enrich-policy-api.html>`_
 
         :arg name: A comma-separated list of enrich policy names
         """
@@ -77,7 +77,7 @@ class EnrichClient(NamespacedClient):
         """
         Creates a new enrich policy.
 
-        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.11/put-enrich-policy-api.html>`_
+        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.12/put-enrich-policy-api.html>`_
 
         :arg name: The name of the enrich policy
         :arg body: The enrich policy to register
@@ -100,7 +100,7 @@ class EnrichClient(NamespacedClient):
         Gets enrich coordinator statistics and information about enrich policies that
         are currently executing.
 
-        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.11/enrich-stats-api.html>`_
+        `<https://www.elastic.co/guide/en/elasticsearch/reference/7.12/enrich-stats-api.html>`_
         """
         return self.transport.perform_request(
             "GET", "/_enrich/_stats", params=params, headers=headers
