@@ -1,22 +1,36 @@
-from operator import methodcaller
-import time
-
-from ..exceptions import TransportError
-from ..compat import map, string_types, Queue
-
-from .errors import ScanError, BulkIndexError
+#  Licensed to Elasticsearch B.V. under one or more contributor
+#  license agreements. See the NOTICE file distributed with
+#  this work for additional information regarding copyright
+#  ownership. Elasticsearch B.V. licenses this file to you under
+#  the Apache License, Version 2.0 (the "License"); you may
+#  not use this file except in compliance with the License.
+#  You may obtain a copy of the License at
+#
+# 	http://www.apache.org/licenses/LICENSE-2.0
+#
+#  Unless required by applicable law or agreed to in writing,
+#  software distributed under the License is distributed on an
+#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+#  KIND, either express or implied.  See the License for the
+#  specific language governing permissions and limitations
+#  under the License.
 
 import logging
+import time
+from operator import methodcaller
 
+from ..compat import Mapping, Queue, map, string_types
+from ..exceptions import TransportError
+from .errors import BulkIndexError, ScanError
 
-logger = logging.getLogger("elasticsearch7.helpers")
+logger = logging.getLogger("elasticsearch.helpers")
 
 
 def expand_action(data):
     """
     From one document or action definition passed in by the user extract the
-    action/data lines needed for elasticsearch7's
-    :meth:`~elasticsearch7.Elasticsearch.bulk` api.
+    action/data lines needed for elasticsearch's
+    :meth:`~elasticsearch.Elasticsearch.bulk` api.
     """
     # when given a string, assume user wants to index raw json
     if isinstance(data, string_types):
@@ -26,9 +40,22 @@ def expand_action(data):
     data = data.copy()
     op_type = data.pop("_op_type", "index")
     action = {op_type: {}}
+
+    # If '_source' is a dict use it for source
+    # otherwise if op_type == 'update' then
+    # '_source' should be in the metadata.
+    if (
+        op_type == "update"
+        and "_source" in data
+        and not isinstance(data["_source"], Mapping)
+    ):
+        action[op_type]["_source"] = data.pop("_source")
+
     for key in (
         "_id",
         "_index",
+        "_if_seq_no",
+        "_if_primary_term",
         "_parent",
         "_percolate",
         "_retry_on_conflict",
@@ -37,6 +64,8 @@ def expand_action(data):
         "_type",
         "_version",
         "_version_type",
+        "if_seq_no",
+        "if_primary_term",
         "parent",
         "pipeline",
         "retry_on_conflict",
@@ -45,13 +74,15 @@ def expand_action(data):
         "version_type",
     ):
         if key in data:
-            if key in [
+            if key in {
+                "_if_seq_no",
+                "_if_primary_term",
                 "_parent",
                 "_retry_on_conflict",
                 "_routing",
                 "_version",
                 "_version_type",
-            ]:
+            }:
                 action[op_type][key[1:]] = data.pop(key)
             else:
                 action[op_type][key] = data.pop(key)
@@ -63,91 +94,77 @@ def expand_action(data):
     return action, data.get("_source", data)
 
 
+class _ActionChunker:
+    def __init__(self, chunk_size, max_chunk_bytes, serializer):
+        self.chunk_size = chunk_size
+        self.max_chunk_bytes = max_chunk_bytes
+        self.serializer = serializer
+
+        self.size = 0
+        self.action_count = 0
+        self.bulk_actions = []
+        self.bulk_data = []
+
+    def feed(self, action, data):
+        ret = None
+        raw_data, raw_action = data, action
+        action = self.serializer.dumps(action)
+        # +1 to account for the trailing new line character
+        cur_size = len(action.encode("utf-8")) + 1
+
+        if data is not None:
+            data = self.serializer.dumps(data)
+            cur_size += len(data.encode("utf-8")) + 1
+
+        # full chunk, send it and start a new one
+        if self.bulk_actions and (
+            self.size + cur_size > self.max_chunk_bytes
+            or self.action_count == self.chunk_size
+        ):
+            ret = (self.bulk_data, self.bulk_actions)
+            self.bulk_actions, self.bulk_data = [], []
+            self.size, self.action_count = 0, 0
+
+        self.bulk_actions.append(action)
+        if data is not None:
+            self.bulk_actions.append(data)
+            self.bulk_data.append((raw_action, raw_data))
+        else:
+            self.bulk_data.append((raw_action,))
+
+        self.size += cur_size
+        self.action_count += 1
+        return ret
+
+    def flush(self):
+        ret = None
+        if self.bulk_actions:
+            ret = (self.bulk_data, self.bulk_actions)
+            self.bulk_actions, self.bulk_data = [], []
+        return ret
+
+
 def _chunk_actions(actions, chunk_size, max_chunk_bytes, serializer):
     """
     Split actions into chunks by number or size, serialize them into strings in
     the process.
     """
-    bulk_actions, bulk_data = [], []
-    size, action_count = 0, 0
+    chunker = _ActionChunker(
+        chunk_size=chunk_size, max_chunk_bytes=max_chunk_bytes, serializer=serializer
+    )
     for action, data in actions:
-        raw_data, raw_action = data, action
-        action = serializer.dumps(action)
-        # +1 to account for the trailing new line character
-        cur_size = len(action.encode("utf-8")) + 1
-
-        if data is not None:
-            data = serializer.dumps(data)
-            cur_size += len(data.encode("utf-8")) + 1
-
-        # full chunk, send it and start a new one
-        if bulk_actions and (
-            size + cur_size > max_chunk_bytes or action_count == chunk_size
-        ):
-            yield bulk_data, bulk_actions
-            bulk_actions, bulk_data = [], []
-            size, action_count = 0, 0
-
-        bulk_actions.append(action)
-        if data is not None:
-            bulk_actions.append(data)
-            bulk_data.append((raw_action, raw_data))
-        else:
-            bulk_data.append((raw_action,))
-
-        size += cur_size
-        action_count += 1
-
-    if bulk_actions:
-        yield bulk_data, bulk_actions
+        ret = chunker.feed(action, data)
+        if ret:
+            yield ret
+    ret = chunker.flush()
+    if ret:
+        yield ret
 
 
-def _process_bulk_chunk(
-    client,
-    bulk_actions,
-    bulk_data,
-    raise_on_exception=True,
-    raise_on_error=True,
-    *args,
-    **kwargs
-):
-    """
-    Send a bulk request to elasticsearch7 and process the output.
-    """
+def _process_bulk_chunk_success(resp, bulk_data, raise_on_error=True):
     # if raise on error is set, we need to collect errors per chunk before raising them
     errors = []
 
-    try:
-        # send the actual request
-        resp = client.bulk("\n".join(bulk_actions) + "\n", *args, **kwargs)
-    except TransportError as e:
-        # default behavior - just propagate exception
-        if raise_on_exception:
-            raise e
-
-        # if we are not propagating, mark all actions in current chunk as failed
-        err_message = str(e)
-        exc_errors = []
-
-        for data in bulk_data:
-            # collect all the information about failed actions
-            op_type, action = data[0].copy().popitem()
-            info = {"error": err_message, "status": e.status_code, "exception": e}
-            if op_type != "delete":
-                info["data"] = data[1]
-            info.update(action)
-            exc_errors.append({op_type: info})
-
-        # emulate standard behavior for failed actions
-        if raise_on_error:
-            raise BulkIndexError(
-                "%i document(s) failed to index." % len(exc_errors), exc_errors
-            )
-        else:
-            for err in exc_errors:
-                yield False, err
-            return
-
     # go through request-response pairs and detect failures
     for data, (op_type, item) in zip(
         bulk_data, map(methodcaller("popitem"), resp["items"])
@@ -168,6 +185,75 @@ def _process_bulk_chunk(
         raise BulkIndexError("%i document(s) failed to index." % len(errors), errors)
 
 
+def _process_bulk_chunk_error(
+    error, bulk_data, raise_on_exception=True, raise_on_error=True
+):
+    # default behavior - just propagate exception
+    if raise_on_exception:
+        raise error
+
+    # if we are not propagating, mark all actions in current chunk as failed
+    err_message = str(error)
+    exc_errors = []
+
+    for data in bulk_data:
+        # collect all the information about failed actions
+        op_type, action = data[0].copy().popitem()
+        info = {"error": err_message, "status": error.status_code, "exception": error}
+        if op_type != "delete":
+            info["data"] = data[1]
+        info.update(action)
+        exc_errors.append({op_type: info})
+
+    # emulate standard behavior for failed actions
+    if raise_on_error:
+        raise BulkIndexError(
+            "%i document(s) failed to index." % len(exc_errors), exc_errors
+        )
+    else:
+        for err in exc_errors:
+            yield False, err
+
+
+def _process_bulk_chunk(
+    client,
+    bulk_actions,
+    bulk_data,
+    raise_on_exception=True,
+    raise_on_error=True,
+    *args,
+    **kwargs
+):
+    """
+    Send a bulk request to elasticsearch and process the output.
+    """
+    kwargs = _add_helper_meta_to_kwargs(kwargs, "bp")
+
+    try:
+        # send the actual request
+        resp = client.bulk("\n".join(bulk_actions) + "\n", *args, **kwargs)
+    except TransportError as e:
+        gen = _process_bulk_chunk_error(
+            error=e,
+            bulk_data=bulk_data,
+            raise_on_exception=raise_on_exception,
+            raise_on_error=raise_on_error,
+        )
+    else:
+        gen = _process_bulk_chunk_success(
+            resp=resp, bulk_data=bulk_data, raise_on_error=raise_on_error
+        )
+    for item in gen:
+        yield item
+
+
+def _add_helper_meta_to_kwargs(kwargs, helper_meta):
+    params = (kwargs or {}).pop("params", {})
+    params["__elastic_client_meta"] = (("h", helper_meta),)
+    kwargs["params"] = params
+    return kwargs
+
+
 def streaming_bulk(
     client,
     actions,
@@ -187,7 +273,7 @@ def streaming_bulk(
     """
     Streaming bulk consumes actions from the iterable passed in and yields
     results per action. For non-streaming usecases use
-    :func:`~elasticsearch7.helpers.bulk` which is a wrapper around streaming
+    :func:`~elasticsearch.helpers.bulk` which is a wrapper around streaming
     bulk that returns summary information about the bulk operation once the
     entire input is consumed and sent.
 
@@ -197,7 +283,7 @@ def streaming_bulk(
     every subsequent rejection for the same chunk, for double the time every
     time up to ``max_backoff`` seconds.
 
-    :arg client: instance of :class:`~elasticsearch7.Elasticsearch` to use
+    :arg client: instance of :class:`~elasticsearch.Elasticsearch` to use
     :arg actions: iterable containing the actions to be executed
     :arg chunk_size: number of docs in one chunk sent to es (default: 500)
     :arg max_chunk_bytes: the maximum size of the request in bytes (default: 100MB)
@@ -274,9 +360,9 @@ def streaming_bulk(
 
 def bulk(client, actions, stats_only=False, *args, **kwargs):
     """
-    Helper for the :meth:`~elasticsearch7.Elasticsearch.bulk` api that provides
+    Helper for the :meth:`~elasticsearch.Elasticsearch.bulk` api that provides
     a more human friendly interface - it consumes an iterator of actions and
-    sends them to elasticsearch7 in chunks. It returns a tuple with summary
+    sends them to elasticsearch in chunks. It returns a tuple with summary
     information - number of successfully executed actions and either list of
     errors or number of errors if ``stats_only`` is set to ``True``. Note that
     by default we raise a ``BulkIndexError`` when we encounter an error so
@@ -286,18 +372,18 @@ def bulk(client, actions, stats_only=False, *args, **kwargs):
     When errors are being collected original document data is included in the
     error dictionary which can lead to an extra high memory usage. If you need
     to process a lot of data and want to ignore/collect errors please consider
-    using the :func:`~elasticsearch7.helpers.streaming_bulk` helper which will
+    using the :func:`~elasticsearch.helpers.streaming_bulk` helper which will
     just return the errors and not store them in memory.
 
 
-    :arg client: instance of :class:`~elasticsearch7.Elasticsearch` to use
+    :arg client: instance of :class:`~elasticsearch.Elasticsearch` to use
     :arg actions: iterator containing the actions
     :arg stats_only: if `True` only report number of successful/failed
         operations instead of just number of successful and a list of error responses
 
     Any additional keyword arguments will be passed to
-    :func:`~elasticsearch7.helpers.streaming_bulk` which is used to execute
-    the operation, see :func:`~elasticsearch7.helpers.streaming_bulk` for more
+    :func:`~elasticsearch.helpers.streaming_bulk` which is used to execute
+    the operation, see :func:`~elasticsearch.helpers.streaming_bulk` for more
     accepted parameters.
     """
     success, failed = 0, 0
@@ -333,7 +419,7 @@ def parallel_bulk(
     """
     Parallel version of the bulk helper run in multiple threads at once.
 
-    :arg client: instance of :class:`~elasticsearch7.Elasticsearch` to use
+    :arg client: instance of :class:`~elasticsearch.Elasticsearch` to use
     :arg actions: iterator containing the actions
     :arg thread_count: size of the threadpool to use for the bulk requests
     :arg chunk_size: number of docs in one chunk sent to es (default: 500)
@@ -356,7 +442,7 @@ def parallel_bulk(
 
     class BlockingPool(ThreadPool):
         def _setup_queues(self):
-            super(BlockingPool, self)._setup_queues()
+            super(BlockingPool, self)._setup_queues()  # type: ignore
             # The queue must be at least the size of the number of threads to
             # prevent hanging when inserting sentinel values during teardown.
             self._inqueue = Queue(max(queue_size, thread_count))
@@ -397,7 +483,7 @@ def scan(
 ):
     """
     Simple abstraction on top of the
-    :meth:`~elasticsearch7.Elasticsearch.scroll` api - a simple iterator that
+    :meth:`~elasticsearch.Elasticsearch.scroll` api - a simple iterator that
     yields all hits as returned by underlining scroll requests.
 
     By default scan does not return results in any pre-determined order. To
@@ -406,8 +492,8 @@ def scan(
     may be an expensive operation and will negate the performance benefits of
     using ``scan``.
 
-    :arg client: instance of :class:`~elasticsearch7.Elasticsearch` to use
-    :arg query: body for the :meth:`~elasticsearch7.Elasticsearch.search` api
+    :arg client: instance of :class:`~elasticsearch.Elasticsearch` to use
+    :arg query: body for the :meth:`~elasticsearch.Elasticsearch.search` api
     :arg scroll: Specify how long a consistent view of the index should be
         maintained for scrolled search
     :arg raise_on_error: raises an exception (``ScanError``) if an error is
@@ -422,10 +508,10 @@ def scan(
         scroll API at the end of the method on completion or error, defaults
         to true.
     :arg scroll_kwargs: additional kwargs to be passed to
-        :meth:`~elasticsearch7.Elasticsearch.scroll`
+        :meth:`~elasticsearch.Elasticsearch.scroll`
 
     Any additional keyword arguments will be passed to the initial
-    :meth:`~elasticsearch7.Elasticsearch.search` call::
+    :meth:`~elasticsearch.Elasticsearch.search` call::
 
         scan(es,
             query={"query": {"match": {"title": "python"}}},
@@ -435,6 +521,7 @@ def scan(
 
     """
     scroll_kwargs = scroll_kwargs or {}
+    _add_helper_meta_to_kwargs(scroll_kwargs, "s")
 
     if not preserve_order:
         query = query.copy() if query else {}
@@ -451,24 +538,28 @@ def scan(
             for hit in resp["hits"]["hits"]:
                 yield hit
 
+            # Default to 0 if the value isn't included in the response
+            shards_successful = resp["_shards"].get("successful", 0)
+            shards_skipped = resp["_shards"].get("skipped", 0)
+            shards_total = resp["_shards"].get("total", 0)
+
             # check if we have any errors
-            if (resp["_shards"]["successful"] + resp["_shards"]["skipped"]) < resp[
-                "_shards"
-            ]["total"]:
+            if (shards_successful + shards_skipped) < shards_total:
+                shards_message = "Scroll request has only succeeded on %d (+%d skipped) shards out of %d."
                 logger.warning(
-                    "Scroll request has only succeeded on %d (+%d skipped) shards out of %d.",
-                    resp["_shards"]["successful"],
-                    resp["_shards"]["skipped"],
-                    resp["_shards"]["total"],
+                    shards_message,
+                    shards_successful,
+                    shards_skipped,
+                    shards_total,
                 )
                 if raise_on_error:
                     raise ScanError(
                         scroll_id,
-                        "Scroll request has only succeeded on %d (+%d skiped) shards out of %d."
+                        shards_message
                         % (
-                            resp["_shards"]["successful"],
-                            resp["_shards"]["skipped"],
-                            resp["_shards"]["total"],
+                            shards_successful,
+                            shards_skipped,
+                            shards_total,
                         ),
                     )
             resp = client.scroll(
@@ -478,7 +569,11 @@ def scan(
 
     finally:
         if scroll_id and clear_scroll:
-            client.clear_scroll(body={"scroll_id": [scroll_id]}, ignore=(404,))
+            client.clear_scroll(
+                body={"scroll_id": [scroll_id]},
+                ignore=(404,),
+                params={"__elastic_client_meta": (("h", "s"),)},
+            )
 
 
 def reindex(
@@ -498,8 +593,8 @@ def reindex(
     to another, potentially (if `target_client` is specified) on a different cluster.
     If you don't specify the query you will reindex all the documents.
 
-    Since ``2.3`` a :meth:`~elasticsearch7.Elasticsearch.reindex` api is
-    available as part of elasticsearch7 itself. It is recommended to use the api
+    Since ``2.3`` a :meth:`~elasticsearch.Elasticsearch.reindex` api is
+    available as part of elasticsearch itself. It is recommended to use the api
     instead of this helper wherever possible. The helper is here mostly for
     backwards compatibility and for situations where more flexibility is
     needed.
@@ -508,20 +603,20 @@ def reindex(
 
         This helper doesn't transfer mappings, just the data.
 
-    :arg client: instance of :class:`~elasticsearch7.Elasticsearch` to use (for
+    :arg client: instance of :class:`~elasticsearch.Elasticsearch` to use (for
         read if `target_client` is specified as well)
     :arg source_index: index (or list of indices) to read documents from
     :arg target_index: name of the index in the target cluster to populate
-    :arg query: body for the :meth:`~elasticsearch7.Elasticsearch.search` api
+    :arg query: body for the :meth:`~elasticsearch.Elasticsearch.search` api
     :arg target_client: optional, is specified will be used for writing (thus
         enabling reindex between clusters)
     :arg chunk_size: number of docs in one chunk sent to es (default: 500)
     :arg scroll: Specify how long a consistent view of the index should be
         maintained for scrolled search
     :arg scan_kwargs: additional kwargs to be passed to
-        :func:`~elasticsearch7.helpers.scan`
+        :func:`~elasticsearch.helpers.scan`
     :arg bulk_kwargs: additional kwargs to be passed to
-        :func:`~elasticsearch7.helpers.bulk`
+        :func:`~elasticsearch.helpers.bulk`
     """
     target_client = client if target_client is None else target_client
     docs = scan(client, query=query, index=source_index, scroll=scroll, **scan_kwargs)
